open Printf;;

type attribute = {
  u : int; 
  s : bool;
  em : bool;
  i : bool;
  b : bool;
  tt : bool;
  sz : int;
  co : char;
};;

type tag =
  | U
  | S
  | EM
  | I
  | B
  | TT
  | PL
  | Sz of int
  | Co of char
;;

type lexeme =
  | Text of char
  | Space of char
  | Open of tag
  | Close of tag
  | EOF
;;

let string_of_tag = function
 | U -> "U"
 | S -> "S"
 | EM -> "EM"
 | I -> "I"
 | B -> "B"
 | TT -> "TT"
 | PL -> "PL"
 | Sz n -> string_of_int n
 | Co c -> String.make 1 c
;;

let outermost = {
  b = false;
  em = false;
  i = false;
  s = false;
  tt = false;
  u = 0; 
  sz = -1;
  co = '_';
};;

type pos = {
  mutable line : int;
  mutable char : int;
  mutable at_eol : bool;
};;

type fileref = {
  mutable ctx : attribute;
  mutable prevspc : attribute option;
  mutable stack : (attribute * tag * int * int) list;
  rest : char Stream.t;
  pos : pos;
  file : string;
};;

let newref s name = {
  ctx = outermost;
  prevspc = None;
  stack = [];
  rest = s;
  pos = {line = 1; char = 0; at_eol = false};
  file = name;
};;

let addchar r =
  if r.pos.at_eol then begin
    r.pos.at_eol <- false;
    r.pos.line <- r.pos.line + 1;
    r.pos.char <- 0;
  end;
  r.pos.char <- r.pos.char + 1;
;;
let newline r =
  addchar r;
  r.pos.at_eol <- true;
;;

exception IllegalChar of fileref * char;;
exception IllegalTag of fileref * char;;
exception TagEOF of fileref;;
exception ExtraClose of fileref * tag;;
exception WrongClose of fileref * tag * int * int * tag;;
exception MissingClose of fileref * tag * int * int;;
type deco_char = char * attribute;;
exception Mismatch of fileref * deco_char * fileref * deco_char;;

let rec tok r = 
  match r.rest with parser
  | [< ''<' >] -> addchar r; lex_tag r
  | [< ' (' ' | '\r' | '\t' as c) >] -> addchar r; Space c
  | [< ''\n' >] -> newline r; Space '\n'
  | [< ' ('\033'..';' | '=' | '?'..'\126' as c) >] -> addchar r; Text c
  | [< 'c >] -> raise (IllegalChar (r, c))
  | [<>] -> EOF

and lex_tag r =
  match r.rest with parser
  | [< ''/' >] -> addchar r; Close (lex_tag_type r)
  | [<>] -> Open (lex_tag_type r)

and lex_tag_type r =
  match r.rest with parser
  | [< ''B' >] -> addchar r; lex_end_tag r; B
  | [< ''E' >] -> addchar r; lex_end_tag2 r 'M'; EM
  | [< ''I' >] -> addchar r; lex_end_tag r; I
  | [< ''P' >] -> addchar r; lex_end_tag2 r 'L'; PL
  | [< ''S' >] -> addchar r; lex_end_tag r; S
  | [< ''T' >] -> addchar r; lex_end_tag2 r 'T'; TT
  | [< ''U' >] -> addchar r; lex_end_tag r; U
  | [< ' ('r'|'g'|'b'|'c'|'m'|'y'|'k'|'w' as c) >] ->
      addchar r; lex_end_tag r; Co c
  | [< ' ('0'..'9' as c) >] ->
      addchar r; lex_end_tag r; Sz (Char.code c - Char.code '0')
  | [< 'c >] -> raise (IllegalTag (r, c))
  | [<>] -> raise (TagEOF r)

and lex_end_tag r =
  match r.rest with parser
  | [< ''>' >] -> addchar r;
  | [< 'c >] -> raise (IllegalTag (r, c))
  | [<>] -> raise (TagEOF r)

and lex_end_tag2 r c =
  match r.rest with parser
  | [< 'cc when cc = c >] -> addchar r; lex_end_tag r;
  | [< 'cc >] -> raise (IllegalTag (r, c))
  | [<>] -> raise (TagEOF r)
;;

let rec next r =
  match tok r with
  | Text c -> r.prevspc <- None; (c, r.ctx)
  | Space c ->
      let spctx = { r.ctx with
        b = false;
        em = false;
        i = false;
        s = false;
        co = if r.ctx.u = 0 then '_' else r.ctx.co
      }
      in
      if r.ctx.tt then (r.prevspc <- None; (c, spctx))
      else if r.prevspc = Some spctx then next r
      else (r.prevspc <- Some spctx; (' ', spctx))
  | Open tag ->
      let nctx =
        match tag with
        | B -> {r.ctx with b = true}
        | EM -> if not r.ctx.s then {r.ctx with em = not r.ctx.em} else r.ctx
        | I -> {r.ctx with i = true}
        | PL -> {r.ctx with b = false; em = false; i = false; s = false;
                 tt = false; u = 0}
        | S -> {r.ctx with s = true; em = false}
        | TT -> {r.ctx with tt = true}
        | U -> if r.ctx.u < 3 then {r.ctx with u = r.ctx.u + 1} else r.ctx
        | Sz n -> {r.ctx with sz = n}
        | Co c -> {r.ctx with co = c}
      in
      let nctx = if nctx = r.ctx then r.ctx else nctx in
      r.stack <- (r.ctx, tag, r.pos.line, r.pos.char) :: r.stack;
      r.ctx <- nctx;
      next r
  | Close ctag ->
      begin match r.stack with
      | [] -> raise (ExtraClose (r, ctag));
      | (ctx, otag, l, c) :: stk when otag <> ctag ->
          raise (WrongClose (r, otag, l, c, ctag));
      | (ctx, otag, l, c) :: stk ->
          r.stack <- stk;
          r.ctx <- ctx;
          next r
      end
  | EOF ->
      begin match r.stack with
      | [] -> ('\000', outermost)
      | (ctx, otag, l, c) :: stk -> raise (MissingClose (r, otag, l, c))
      end
;;

let compare r1 c1 r2 c2 = if c1 <> c2 then raise (Mismatch (r1, c1, r2, c2));;

let cannot_open filename errmsg =
  fprintf stderr "Cannot open \"%s\": %s\n" filename errmsg
;;

let char_printable c =
  match c with
  | '\009' -> "TAB"
  | '\010' -> "LF"
  | '\013' -> "CR"
  | '\032' -> "SPC"
  | '\033'..'\126' -> sprintf "\"%c\"" c
  | _ -> sprintf "0x%02x" (Char.code c)
;;

let string_of_deco_char (c, ctx) =
  let attr = (if ctx.b then ",B" else "")
             ^ (if ctx.em then ",EM" else "")
             ^ (if ctx.i then ",I" else "")
             ^ (if ctx.s then ",S" else "")
             ^ (if ctx.tt then ",TT" else "")
             ^ (if ctx.u > 0 then ",U="^(string_of_int ctx.u) else "")
             ^ (if ctx.sz <> -1 then ","^(string_of_int ctx.sz) else "")
             ^ (if ctx.co <> '_' then ","^(String.make 1 ctx.co) else "")
  in
  if c = '\000' then begin
    "[end-of-file]"
  end else if String.length attr > 0 then begin
    attr.[0] <- '{';
    (char_printable c) ^ attr ^ "}"
  end else begin
    (char_printable c)
  end
;;

let printlcpos r l c =
  let p = if c = 0 then 0 else c - 1 in
  match Sys.os_type with
  | "MacOS" ->
      fprintf stderr "file \"%s\"; line %d; characters %d to %d\n" r.file l p c
  | _ ->
      fprintf stderr "File \"%s\", line %d, characters %d-%d:\n" r.file l p c
;;

let printpos r = printlcpos r r.pos.line r.pos.char;;

let report_error e =
  begin match e with
  | IllegalChar (r, c) ->
      printpos r;
      fprintf stderr "### illegal character: %s\n\n" (char_printable c);
  | IllegalTag (r, c) ->
      printpos r;
      fprintf stderr "### bad character in tag name: %s\n\n" (char_printable c);
  | TagEOF r ->
      printpos r;
      fprintf stderr "### end-of-file in the middle of a tag\n\n";
  | ExtraClose (r, t) ->
      printpos r;
      fprintf stderr "### this tag was never opened: </%s>\n\n"
              (string_of_tag t);
  | WrongClose (r, ot, l, c, ct) ->
      printlcpos r l c;
      printpos r;
      fprintf stderr "### <%s> is closed by </%s>\n\n" (string_of_tag ot)
              (string_of_tag ct);
  | MissingClose (r, t, l, c) ->
      printlcpos r l c;
      fprintf stderr "### this tag is never closed: <%s>\n\n" (string_of_tag t);
  | Mismatch (r1, dc1, r2, dc2) ->
      printpos r1;
      printpos r2;
      fprintf stderr "### mismatch between %s (in %s) and %s (in %s)\n\n"
              (string_of_deco_char dc1) r1.file
              (string_of_deco_char dc2) r2.file;
  | _ -> raise e;
  end;
  flush stderr;
;;

let fatal errcode =
  fprintf stderr "### cannot continue (error in reference file)\n";
  exit errcode;
;;

let main () =
  if Array.length Sys.argv < 2 || Sys.argv.(1) = "-help"
     || Sys.argv.(1) = "--help"
  then begin
    fprintf stderr "usage: check <reffile> [<file_1>...<file_n>]\n";
    exit 0;
  end;
  let f i = if i < 2 then None
            else try let s = Stream.of_channel (open_in Sys.argv.(i)) in
                     Some (newref s Sys.argv.(i))
                 with Sys_error e -> (cannot_open Sys.argv.(i) e; None)
  in
  let reffile =
    try newref (Stream.of_channel (open_in Sys.argv.(1))) Sys.argv.(1)
    with Sys_error e -> (cannot_open Sys.argv.(1) e; fatal 3)
  in
  let files = Array.init (Array.length Sys.argv) f in
  begin try
    while true do
      let c = next reffile in
      for i = 2 to Array.length files - 1 do
        match files.(i) with
        | None -> ()
        | Some r ->
            begin try compare reffile c r (next r) with
            | e ->
                report_error e;
                files.(i) <- None;
            end
      done;
      if c = ('\000', outermost) then raise Exit;
    done;
  with
    | Exit -> ()
    | e -> report_error e; fatal 4;
  end;
  let nbeq = ref 0 in
  for i = 2 to Array.length files - 1 do
    match files.(i) with
    | None -> ()
    | Some r -> incr nbeq;
  done;
  if !nbeq > 0 then begin
    fprintf stderr "%d files are equivalent to %s :\n" !nbeq reffile.file;
    flush stderr;
    for i = 2 to Array.length files - 1 do
      match files.(i) with
      | None -> ()
      | Some r -> printf "%s\n" r.file;
    done;
    flush stdout;
    exit 0;
  end else if Array.length Sys.argv = 2 then begin
    exit 0;
  end else begin
    exit 4;
  end;
;;

main ();;
